import abc
import os
import pickle
import subprocess
import types
from io import BufferedReader
from typing import Any, Callable, Dict, List, TypeVar, Union

import cloudpickle
import uvicorn
import yaml
from fastapi import Depends, FastAPI, HTTPException, Request
from schema import Or, Schema

from vessl.cli._util import Endpoint, format_url
from vessl.model import (
    create_model,
    download_model_volume_file,
    list_model_volume_files,
    read_model,
    upload_model_volume_file,
)
from vessl.organization import _get_organization_name
from vessl.util import logger
from vessl.util.common import get_module
from vessl.util.exception import (
    InvalidParamsError,
    VesslApiException,
    VesslRuntimeException,
)

ModelType = TypeVar("ModelType")
InputDataType = bytes
ModelInputDataType = TypeVar("ModelInputDataType")
ModelOutputDataType = TypeVar("ModelOutputDataType")
OutputDataType = TypeVar("OutputDataType")

PROTOCOL_VERSION = 4
USER_DEFINED = "user-defined"
MANIFEST_FILENAME = "vessl.manifest.yaml"
RUNNER_FILENAME = "vessl.runner.pkl"
DEFAULT_MODEL_FILENAME = "vessl.model.pkl"


class RunnerBase(abc.ABC):
    """Base class for model registering.

    This base class introduces 5 static methods as followings:
    - `predict`: Make prediction with given data and model. This method must be overridden. The
      data is given from the result of `preprocess_data`, and the return value of this method
      will be passed to `postprocess_data` before serving.
    - `save_model`: Save the model into a file. Return value of this method will be given to the
      `load_model` method on model loading. If this method is overriden, `load_model` must be
      overriden as well.
    - `load_model`: Load the model from a file.
    - `preprocess_data`: Preprocess the data before prediction. It converts the API input data to
      the model input data.
    - `postprocess_data`: Postprocess the data after prediction. It converts the model output data
      to the API output data.

    Check each method's docstring for more information.
    """

    @staticmethod
    def save_model(model: ModelType) -> Dict[str, str]:
        """Save the given model instance into file.

        Return value of this method will be given to first argument of `load_model` on model loading.

        Args:
            model(model_instance): Model instance to save.

        Returns:
            (dict) Data that will be passed to `load_model` on model loading.
                Must be a dictionary with key and value both string.
        """
        pass

    @staticmethod
    def load_model(props: Union[Dict[str, str], None], artifacts: Dict[str, str]) -> ModelType:
        """Load the model instance from file.

        `props` is given from the return value of `save_model`, and `artifacts` is
        given from the `register_model` method.

        If the `save_model` is not overriden, `props` will be None

        Args:
            props(dict | None): Data that was returned by `save_model`. If `save_model` is
                not overriden, this will be None.
            artifacts(dict): Data that is given by `register_model` function.

        Returns:
            Model instance.
        """
        pass

    @staticmethod
    def preprocess_data(data: InputDataType) -> ModelInputDataType:
        """Preprocess the given data.

        The data processed by this method will be given to the model.

        Args:
            data: Data to be preprocessed.

        Returns:
            Preprocessed data that will be given to the model.
        """
        pass

    @staticmethod
    def postprocess_data(data: ModelOutputDataType) -> OutputDataType:
        """Postprocess the given data.

        The data processed by this method will be given to the user.

        Args:
            data: Data to be postprocessed.

        Returns:
            Postprocessed data that will be given to the user.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def predict(model: ModelType, data: ModelInputDataType) -> ModelOutputDataType:
        """Make prediction with given data and model.

        Args:
            model(model_instance): Model instance.
            data: Data to be predicted.

        Returns:
            Prediction result.
        """
        raise NotImplementedError()


_manifest_schema = Schema(
    {
        "version": str,
        "model": {
            "save_type": Or(str, None),
            "save_path": Or(str, None),
            "save_output": Or({str: str}, None),
        },
        "runner": {
            "save_type": str,
            "save_path": str,
        },
        "use_default": [str],
        "requirements": [str],
        "artifacts": Or({str: str}, {}),
    }
)


def _save_object(save_type: Union[str, None], obj: Any, path: str) -> str:
    """Save the object into `path`

    If `save_type` is not `None`, use the given serialization method. It will not fallback on failure.
    Otherwise, use various methods to serialize the object as best as possible.
    After serialization, save it into `path`.

    Args:
        save_type (str | None): The serialization method.
        obj (Any): The object to serialize.
        path (str): The path to save the object.

    Returns:
        (str) The save type finally applied to the object.
    """
    # Known save types
    if save_type == "torch":
        import torch

        with open(path, "wb") as f:
            torch.save(obj, f, pickle_module=cloudpickle, pickle_protocol=PROTOCOL_VERSION)
        return "torch"
    if save_type is not None:
        # save_type given, but we don't know how to handle it
        raise InvalidParamsError("Unknown save type: {}".format(save_type))

    # Fallback saving methods for None save_type
    with open(path, "wb") as f:
        # Try cloudpickle first
        try:
            cloudpickle.dump(obj, f, protocol=PROTOCOL_VERSION)
            return "cloudpickle"
        except pickle.PicklingError:
            pass

        # Try joblib
        try:
            joblib = get_module("joblib")
            joblib.dump(obj, f)
            return "joblib"
        except pickle.PicklingError:
            pass

        # Last fallback. Use pickle module
        try:
            pickle.dump(obj, f, protocol=PROTOCOL_VERSION)
            return "pickle"
        except pickle.PicklingError:
            pass
    # Failed to save. Raise.
    raise VesslRuntimeException("Could not save the object")


def _try_load(loader: Callable, br: BufferedReader) -> Any:
    try:
        return loader(br)
    except Exception as e:
        raise VesslRuntimeException(
            f"Could not load the object. Did you use a different Python version to register the model?\nError: {e}"
        )


def _load_object(save_type: str, path: str) -> Any:
    """Load the object from given path

    It uses the given save_type to find a correct way to deserialize the object.

    Args:
        save_type (str): The serialization method used to save the object.
        path (str): The path to load the object.

    Returns:
        (Any) The loaded object.
    """
    if save_type == "torch":
        with open(path, "rb") as f:
            torch = get_module(
                "torch", required='torch package is required. Run "pip install torch".'
            )
            return _try_load(torch.load, f)

    with open(path, "rb") as f:
        if save_type == "cloudpickle":
            return _try_load(cloudpickle.load, f)
        if save_type == "joblib":
            joblib = get_module(
                "joblib",
                required='joblib package is required. Run "pip install joblib".',
            )
            return _try_load(joblib.load, f)
        if save_type == "pickle":
            return _try_load(pickle.load, f)
    raise InvalidParamsError("Unknown save type")


def _upload_artifacts(
    artifacts: Dict[str, str],
    *,
    repository_name: str,
    model_number: int,
    organization_name: str,
) -> None:
    """Upload given artifacts to the model volume.

    Args:
        artifacts(dict): Artifacts to be uploaded. Key is the path to artifact in local filesystem, and value is the
            path in the model volume. Only trailing asterisk(*) is allowed for glob pattern.
        repository_name(str): Name of the repository.
        model_number(int): Number of the model.
        organization_name(str): Name of the organization.

    Example:
        >>> _upload_artifacts({
        >>>     "./output/*": "*",
        >>>     "./sample.data": "sample.data",
        >>> }, repository_name="my-repository", model_number=1, organization_name="my-organization")
    """
    # Input Validation
    for local_path, remote_path in artifacts.items():
        local_dir, local_base = os.path.dirname(local_path), os.path.basename(local_path)
        remote_dir, remote_base = os.path.dirname(remote_path), os.path.basename(remote_path)
        if "*" in local_dir or "*" in remote_dir:
            raise InvalidParamsError("Only trailing asterisk(*) is allowed for glob pattern")
        if ("*" in local_base and local_base != "*") or ("*" in remote_base and remote_base != "*"):
            raise InvalidParamsError("Only trailing asterisk(*) is allowed for glob pattern")
        if (local_base == "*") != (remote_base == "*"):
            # Both should either or neither be glob pattern
            raise InvalidParamsError(
                "Both local and remote path should either or neither be glob pattern"
            )

    if len(artifacts.keys()) != 0:
        print("Artifacts will be uploaded as follows:")
        for l, r in artifacts.items():
            print("  {} -> {}".format(l, r))

    for local_path, remote_path in artifacts.items():
        local_dir, local_base = os.path.dirname(local_path), os.path.basename(local_path)
        remote_dir, remote_base = os.path.dirname(remote_path), os.path.basename(remote_path)
        if local_base == "*":
            # Glob pattern
            for root, _, files in os.walk(local_dir):
                for filename in files:
                    local_filepath = os.path.join(root, filename)
                    upload_model_volume_file(
                        repository_name=repository_name,
                        model_number=model_number,
                        organization_name=organization_name,
                        source_path=local_filepath,
                        dest_path=local_filepath.replace(local_dir, remote_dir),
                    )
        else:
            # Upload single file
            upload_model_volume_file(
                repository_name=repository_name,
                model_number=model_number,
                organization_name=organization_name,
                source_path=local_path,
                dest_path=remote_path,
            )


# Related to model registration. Will be used as Python SDK, not CLI.
def _register_serving_method(
    organization_name: str,
    repository_name: str,
    model_number: Union[int, None],
    model_type: Union[str, None],
    model_instance: Union[ModelType, None],
    runner_cls: RunnerBase,
    requirements: List[str],
    artifacts: Dict[str, str],
):
    # Argument validation
    # Valid comb of (model_number, model_type, model_instance) (0 means None)
    # - (0, 0, 1): Create new model from given model_instance and register. Default fallback saver will be used since model_type is not given.
    # - (0, 1, 1): Create new model from given model_instance and register. model_type's saver will be used.
    # - (1, 0, 0): Register existing model. It will only add runner/manifest file to make model servable.
    # - (1, 0, 1): Overwrite existing model. Given model_instance will be serialized and saved to given model number. Default fallback saver will be used since model_type is not given.
    # - (1, 1, 1): Overwrite existing model. Given model_instance will be serialized and saved to given model number. model_type's saver will be used.
    if model_number is None and model_instance is None:  # Exclude (0, x, 0)
        raise InvalidParamsError("`model_instance` must be given to create a new model")
    if model_instance is None and model_type is not None:  # Exclude (x, 1, 0)
        raise InvalidParamsError("`model_instance` must be given if `model_type` is not None")
    for key in artifacts:
        if key in [RUNNER_FILENAME, MANIFEST_FILENAME, DEFAULT_MODEL_FILENAME]:
            raise InvalidParamsError("`{}` is a reserved key of artifacts".format(key))

    if model_number is not None:
        # Check model existence
        try:
            read_model(
                repository_name=repository_name,
                model_number=model_number,
                organization_name=organization_name,
            )
        except VesslApiException:
            raise InvalidParamsError("Failed to read requested model")
    else:
        # Create new model
        logger.info("`model_number` is None. New model will be created.")
        res = create_model(
            repository_name=repository_name,
            organization_name=organization_name,
        )
        model_number = res.number

    # Runner class validation
    runner_cls()  # Emits error if abstract method is not overriden
    if model_instance is None:
        # Tries to add serving method only.
        if runner_cls.load_model == RunnerBase.load_model:
            raise InvalidParamsError(
                "`load_model` method must be overriden if model_instance is not given"
            )
    else:
        if (
            runner_cls.save_model != RunnerBase.save_model
            and runner_cls.load_model == RunnerBase.load_model
        ):
            raise InvalidParamsError(
                "`load_model` must be implemented if `save_model` is implemented"
            )
    # Validation done.

    # Business logic from here
    # Save the model
    model_save_type, model_save_path, model_save_output = USER_DEFINED, None, None
    if model_instance is not None:
        if runner_cls.save_model != RunnerBase.save_model:
            # Use the custom `save_model` if defined
            model_save_type = USER_DEFINED
            model_save_output = runner_cls.save_model(model_instance)
        elif runner_cls.load_model != RunnerBase.load_model:
            # `save_model` is not overriden but `load_model` is overridden.
            # We can assume user saves the model in other way. Skip saving
            model_save_type = USER_DEFINED
        else:
            # Use the default saver
            model_save_type = _save_object(model_type, model_instance, DEFAULT_MODEL_FILENAME)
            model_save_path = DEFAULT_MODEL_FILENAME

    # Save the runner
    runner_save_type, runner_save_path = "", ""
    runner_save_type = _save_object(None, runner_cls, RUNNER_FILENAME)
    runner_save_path = RUNNER_FILENAME

    # Mark the methods that are not overriden by user
    use_default = []
    if runner_cls.load_model == RunnerBase.load_model:
        use_default.append("load_model")
    if runner_cls.preprocess_data == RunnerBase.preprocess_data:
        use_default.append("preprocess_data")
    if runner_cls.postprocess_data == RunnerBase.postprocess_data:
        use_default.append("postprocess_data")

    # Save the manifest
    manifest = {
        "version": "v1",
        "model": {
            "save_type": model_save_type,
            "save_path": model_save_path,
            "save_output": model_save_output,
        },
        "runner": {
            "save_type": runner_save_type,
            "save_path": runner_save_path,
        },
        "use_default": use_default,
        "requirements": requirements,
        "artifacts": artifacts,
    }
    _manifest_schema.validate(manifest)
    with open(MANIFEST_FILENAME, "w") as f:
        yaml.dump(manifest, f)

    # Upload artifacts
    upload = artifacts.copy()
    upload[MANIFEST_FILENAME] = MANIFEST_FILENAME
    upload[runner_save_path] = runner_save_path
    if model_save_path is not None:
        upload[model_save_path] = model_save_path
    _upload_artifacts(
        upload,
        repository_name=repository_name,
        model_number=model_number,
        organization_name=organization_name,
    )

    print(
        f"Successfully registered model: {format_url(Endpoint.model.format(organization_name, repository_name, model_number))}"
    )


def register_model(
    repository_name: str,
    model_number: Union[int, None],
    runner_cls: RunnerBase,
    model_instance: Union[ModelType, None] = None,
    requirements: List[str] = None,
    artifacts: Dict[str, str] = None,
    **kwargs,
):
    """Register the given model for serving. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        repository_name(str): Model repository name.
        model_number(int | None): Model number. If None, new model will be
            created. In such case, `model_instance` must be given.
        runner_cls(RunnerBase): Runner class that includes code for serving.
        model_instance(ModelType | None): Model instance. If None, `runner_cls`
            must override `load_model` method. Defaults to None.
        requirements(List[str]): Python requirements for the model. Defaults to
            [].
        artifacts(Dict[str, str]): Artifacts to be uploaded. Key is the path to
            artifact in local filesystem, and value is the path in the model
            volume. Only trailing asterisk(*) is allowed for glob pattern.
            Defaults to {}.

    Example:
        ```python
        register_model(
            repository_name="my-model",
            model_number=1,
            runner_cls=MyRunner,
            model_instance=model_instance,
            requirements=["torch", "torchvision"],
            artifacts={"model.pt": "model.pt", "checkpoints/*": "checkpoints/*"},
        )
        ```
    """
    organization_name = _get_organization_name(**kwargs)
    if requirements is None:
        requirements = []
    if artifacts is None:
        artifacts = {}
    _register_serving_method(
        organization_name=organization_name,
        repository_name=repository_name,
        model_number=model_number,
        model_type=None,
        model_instance=model_instance,
        runner_cls=runner_cls,
        requirements=requirements,
        artifacts=artifacts,
    )


def register_torch_model(
    repository_name: str,
    model_number: Union[int, None],
    model_instance: ModelType,
    preprocess_data=None,
    postprocess_data=None,
    requirements: List[str] = None,
    **kwargs,
):
    """Register the given torch model instance for model serving. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        repository_name(str): Model repository name.
        model_number(int | None): Model number. If None, new model will be
            created.
        model_instance(model_instance): Torch model instance.
        preprocess_data(callable): Function that will preprocess data.
            Defaults to identity function.
        postprocess_data(callable): Function that will postprocess data.
            Defaults to identity function.
        requirements(list): List of requirements. Defaults to [].

    Example:
        ```python
        vessl.register_model(
            repository_name="my-model",
            model_number=1,
            model_instance=model_instance,
            preprocess_data=lambda x: int(x),
            postprocess_data=lambda x: {"prediction": x},
            requirements=["torch", "torchvision"],
        )
        ```
    """
    # Check if the model is torch
    torch_nn = get_module(
        "torch.nn", required='torch package is required. Run "pip install torch".'
    )
    if not isinstance(model_instance, torch_nn.Module):
        raise InvalidParamsError("model_instance must be a torch.nn.Module.")

    organization_name = _get_organization_name(**kwargs)
    if requirements is None:
        requirements = []

    class PyTorchRunner(RunnerBase):
        @staticmethod
        def predict(model: ModelType, data: ModelInputDataType) -> ModelOutputDataType:
            return model(data)

    if preprocess_data is not None:
        PyTorchRunner.preprocess_data = preprocess_data
    if postprocess_data is not None:
        PyTorchRunner.postprocess_data = postprocess_data
    _register_serving_method(
        organization_name=organization_name,
        repository_name=repository_name,
        model_number=model_number,
        model_type="torch",
        model_instance=model_instance,
        runner_cls=PyTorchRunner,
        requirements=requirements + ["torch", "torchvision"],
        artifacts={},
    )


# Related to model serving. Will be used by CLI.
class _ModelServer(object):
    """Model server class."""

    def __init__(self, model: ModelType, preprocess_data, predict, postprocess_data):
        self.model = model
        self.preprocess_data = preprocess_data
        self.predict = predict
        self.postprocess_data = postprocess_data

    def run(self, data: InputDataType) -> OutputDataType:
        processed = self.preprocess_data(data)
        result = self.predict(self.model, processed)
        return self.postprocess_data(result)


def _load_model(install_reqs: bool = False) -> _ModelServer:
    """Load model from file. This function assumes that cwd is the root of model volume.

    Args:
        install_reqs(bool): Whether to install requirements. Default is False.

    Returns:
        (ModelServer) Model server.
    """
    with open(MANIFEST_FILENAME, "r") as f:
        try:
            manifest = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.YAMLError:
            raise VesslRuntimeException("Invalid manifest file. Please register the model again.")

    _manifest_schema.validate(manifest)
    # Check for requirements
    if install_reqs:
        reqs = manifest["requirements"]
        if reqs:
            logger.info("Installing requirements:", reqs)
            subprocess.check_call(["pip", "install", *reqs])  # Support for conda?

    # Load runner
    runner_cls = _load_object(manifest["runner"]["save_type"], manifest["runner"]["save_path"])
    # Load model
    if "load_model" in manifest["use_default"]:
        model = _load_object(manifest["model"]["save_type"], manifest["model"]["save_path"])
    else:  # Use custom loader
        model = runner_cls.load_model(manifest["model"]["save_output"], manifest["artifacts"])
    # Load methods
    preprocess_data = (
        (lambda x: x)
        if "preprocess_data" in manifest["use_default"]
        else runner_cls.preprocess_data
    )
    postprocess_data = (
        (lambda x: x)
        if "postprocess_data" in manifest["use_default"]
        else runner_cls.postprocess_data
    )
    model_server = _ModelServer(model, preprocess_data, runner_cls.predict, postprocess_data)
    return model_server


def _run_model_server(model: _ModelServer, remote: bool):
    """Boot simple server that serves the model."""
    app = FastAPI()

    def verify_key(request: Request):
        if not remote:
            return True

        token = request.headers.get("X-AUTH-KEY")
        if token is None or token != os.environ.get("SERVING_AUTH_KEY"):
            raise HTTPException(status_code=401, detail="Invalid auth key.")
        return True

    @app.post("/")
    async def index(request: Request, authorized: bool = Depends(verify_key)):
        if authorized:
            body = await request.body()
            print("Received:", body)
            result = model.run(body)
            print("Result:", result)
            return result

    uvicorn.run(app, host="0.0.0.0", port=8000)


def serve_model(
    repository_name: str,
    model_number: int,
    install_reqs: bool = False,
    remote: bool = False,
    **kwargs,
):
    """Execute the server which serves the given model. If you want to override the default organization, then pass `organization_name` as `**kwargs`.

    This method will create `.vessl-serving` under current directory to save the artifacts in local.

    Args:
        repository_name(str): Model repository name.
        model_number(int): Model number.
        install_reqs(bool): Whether to install requirements. Defaults to False.
    """
    # Validation
    organization_name = _get_organization_name(**kwargs)
    model = read_model(repository_name=repository_name, model_number=model_number, **kwargs)
    if not model.is_servable:
        raise InvalidParamsError("Model is not servable.")

    if not remote:
        print("Downloading model artifacts for serving...")
        # Create temporal directory if not exists
        local_path_root = f".vessl-serving/{organization_name}/{repository_name}/{model_number}"
        os.makedirs(local_path_root, exist_ok=True)

        # Get the list of files to be downloaded. Only the outdated files or the files that are not in the local will be downloaded.
        volume_files = list_model_volume_files(
            repository_name=repository_name, model_number=model_number, recursive=True
        )
        local_status_filename = (
            ".vessl.localstatus.yaml"  # File that keeps track of the local status of the files
        )
        local_status = {  # Structure of the local status file
            "hash": {},
        }
        download_list = []
        if os.path.exists(os.path.join(local_path_root, local_status_filename)):
            # Local status file exists -> do not download the files that are already up-to-date.
            with open(os.path.join(local_path_root, local_status_filename), "r") as f:
                local_status.update(yaml.load(f, Loader=yaml.FullLoader))
            for volume_file in volume_files:
                if volume_file.is_dir:
                    continue  # Skip directory
                if volume_file.hash == local_status.get("hash", {}).get(volume_file.path, None):
                    continue  # Skip if hash is up to date
                download_list.append(volume_file)
        else:
            # Local status file does not exist -> download all files.
            for volume_file in volume_files:
                if volume_file.is_dir:
                    continue
                download_list.append(volume_file)
        # Download list retrieved.

        # Download files
        os.chdir(local_path_root)  # Make root of the model volume as cwd
        for volume_file in download_list:
            download_model_volume_file(
                repository_name=repository_name,
                model_number=model_number,
                source_path=volume_file.path,
                dest_path=volume_file.path,
            )
            # Update hash so that we don't download the same file again.
            local_status["hash"][volume_file.path] = volume_file.hash
        with open(local_status_filename, "w") as f:
            yaml.dump(local_status, f)
        print("Artifacts for model serving are up to date.")

    # Serve model
    print("Loading model from artifacts...")
    model = _load_model(install_reqs)
    print("Booting local server for serving...")
    _run_model_server(model, remote)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class PipelineContextMetadata(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'current_step_id': 'int',
        'current_step_type': 'str',
        'pipeline_arguments': 'dict[str, str]',
        'pipeline_execution_id': 'int',
        'step_key_dictionary': 'dict[str, int]'
    }

    attribute_map = {
        'current_step_id': 'current_step_id',
        'current_step_type': 'current_step_type',
        'pipeline_arguments': 'pipeline_arguments',
        'pipeline_execution_id': 'pipeline_execution_id',
        'step_key_dictionary': 'step_key_dictionary'
    }

    def __init__(self, current_step_id=None, current_step_type=None, pipeline_arguments=None, pipeline_execution_id=None, step_key_dictionary=None, local_vars_configuration=None):  # noqa: E501
        """PipelineContextMetadata - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._current_step_id = None
        self._current_step_type = None
        self._pipeline_arguments = None
        self._pipeline_execution_id = None
        self._step_key_dictionary = None
        self.discriminator = None

        if current_step_id is not None:
            self.current_step_id = current_step_id
        if current_step_type is not None:
            self.current_step_type = current_step_type
        if pipeline_arguments is not None:
            self.pipeline_arguments = pipeline_arguments
        if pipeline_execution_id is not None:
            self.pipeline_execution_id = pipeline_execution_id
        if step_key_dictionary is not None:
            self.step_key_dictionary = step_key_dictionary

    @property
    def current_step_id(self):
        """Gets the current_step_id of this PipelineContextMetadata.  # noqa: E501


        :return: The current_step_id of this PipelineContextMetadata.  # noqa: E501
        :rtype: int
        """
        return self._current_step_id

    @current_step_id.setter
    def current_step_id(self, current_step_id):
        """Sets the current_step_id of this PipelineContextMetadata.


        :param current_step_id: The current_step_id of this PipelineContextMetadata.  # noqa: E501
        :type current_step_id: int
        """

        self._current_step_id = current_step_id

    @property
    def current_step_type(self):
        """Gets the current_step_type of this PipelineContextMetadata.  # noqa: E501


        :return: The current_step_type of this PipelineContextMetadata.  # noqa: E501
        :rtype: str
        """
        return self._current_step_type

    @current_step_type.setter
    def current_step_type(self, current_step_type):
        """Sets the current_step_type of this PipelineContextMetadata.


        :param current_step_type: The current_step_type of this PipelineContextMetadata.  # noqa: E501
        :type current_step_type: str
        """

        self._current_step_type = current_step_type

    @property
    def pipeline_arguments(self):
        """Gets the pipeline_arguments of this PipelineContextMetadata.  # noqa: E501


        :return: The pipeline_arguments of this PipelineContextMetadata.  # noqa: E501
        :rtype: dict[str, str]
        """
        return self._pipeline_arguments

    @pipeline_arguments.setter
    def pipeline_arguments(self, pipeline_arguments):
        """Sets the pipeline_arguments of this PipelineContextMetadata.


        :param pipeline_arguments: The pipeline_arguments of this PipelineContextMetadata.  # noqa: E501
        :type pipeline_arguments: dict[str, str]
        """

        self._pipeline_arguments = pipeline_arguments

    @property
    def pipeline_execution_id(self):
        """Gets the pipeline_execution_id of this PipelineContextMetadata.  # noqa: E501


        :return: The pipeline_execution_id of this PipelineContextMetadata.  # noqa: E501
        :rtype: int
        """
        return self._pipeline_execution_id

    @pipeline_execution_id.setter
    def pipeline_execution_id(self, pipeline_execution_id):
        """Sets the pipeline_execution_id of this PipelineContextMetadata.


        :param pipeline_execution_id: The pipeline_execution_id of this PipelineContextMetadata.  # noqa: E501
        :type pipeline_execution_id: int
        """

        self._pipeline_execution_id = pipeline_execution_id

    @property
    def step_key_dictionary(self):
        """Gets the step_key_dictionary of this PipelineContextMetadata.  # noqa: E501


        :return: The step_key_dictionary of this PipelineContextMetadata.  # noqa: E501
        :rtype: dict[str, int]
        """
        return self._step_key_dictionary

    @step_key_dictionary.setter
    def step_key_dictionary(self, step_key_dictionary):
        """Sets the step_key_dictionary of this PipelineContextMetadata.


        :param step_key_dictionary: The step_key_dictionary of this PipelineContextMetadata.  # noqa: E501
        :type step_key_dictionary: dict[str, int]
        """

        self._step_key_dictionary = step_key_dictionary

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PipelineContextMetadata):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PipelineContextMetadata):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import Dict, List, Optional

from openapi_client import OrmVolumeMountRequestSourceModelVolume
from openapi_client.models import (
    OrmVolumeMountRequest,
    OrmVolumeMountRequests,
    OrmVolumeMountRequestSourceArchiveFile,
    OrmVolumeMountRequestSourceCode,
    OrmVolumeMountRequestSourceDataset,
    OrmVolumeMountRequestSourceDatasetVersion,
    OrmVolumeMountRequestSourceObjectStorage,
    StorageFile,
    VolumeFileCopyAPIInput,
    VolumeFileCreateAPIInput,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.project import _get_project
from vessl.util.constant import (
    DATASET_VERSION_HASH_LATEST,
    EXPERIMENT_WORKING_DIR,
    MOUNT_PATH_EMPTY_DIR,
    SOURCE_TYPE_ARCHIVE_FILE,
    SOURCE_TYPE_OBJECT_STORAGE,
    SOURCE_TYPE_CODE,
    SOURCE_TYPE_DATASET,
    SOURCE_TYPE_DATASET_VERSION,
    SOURCE_TYPE_EMPTY_DIR,
    SOURCE_TYPE_MODEL_VOLUME,
    SOURCE_TYPE_OUTPUT,
    SOURCE_TYPE_OBJECT_STORAGE,
    TEMP_DIR,
)
from vessl.util.downloader import Downloader
from vessl.util.exception import (
    GitError,
    InvalidParamsError,
    InvalidVolumeFileError,
    VesslApiException,
)
from vessl.util.git_local import get_git_diff_path, get_git_ref, get_git_repo
from vessl.util.random import random_string
from vessl.util.tar import Tar
from vessl.util.uploader import Uploader
from vessl.util.volume import VolumeFileTransfer


def read_volume_file(volume_id: int, path: str) -> StorageFile:
    """Read a file in the volume.

    Args:
        volume_id(int): Volume ID.
        path(str): Path within the volume.

    Example:
        ```python
        vessl.read_volume_file(
            volume_id=123456,
            path="train.csv",
        )
        ```
    """
    return vessl_api.volume_file_read_api(volume_id=volume_id, path=path)


def list_volume_files(
    volume_id: int,
    need_download_url: bool = False,
    path: str = "",
    recursive: bool = False,
) -> List[StorageFile]:
    """List files in the volume.

    Args:
        volume_id(int): Volume ID.
        need_download_url(bool): True if you need a download URL, False
            otherwise. Defaults to False.
        path(str): Path within the volume. Defaults to root.
        recursive(bool): True if list files recursively, False otherwise.
            Defaults to False.

    Example:
        ```python
        vessl.list_volume_files(
            volume_id=123456,
        )
        ```
    """
    try:
        result = vessl_api.volume_file_list_api(
            volume_id=volume_id,
            recursive=recursive,
            path=path,
            need_download_url=need_download_url,
        ).results
    except VesslApiException as e:
        if path == "":
            result = []
        else:
            raise e

    return sorted(result, key=lambda x: x.path)


def create_volume_file(volume_id: int, is_dir: bool, path: str) -> StorageFile:
    """Create file in the volume.

    Args:
        volume_id(int): Volume ID.
        is_dir(bool): True if a file is directory, False otherwise.
        path(str): Path within the volume.

    Example:
        ```python
        vessl.create_volume_file(
            volume_id=123456,
            is_dir=False,
            path="models"
        )
        ```
    """
    result = vessl_api.volume_file_create_api(
        volume_id=volume_id,
        volume_file_create_api_input=VolumeFileCreateAPIInput(
            is_dir=is_dir,
            path=path,
        ),
    )
    assert isinstance(result, StorageFile)
    return result


def delete_volume_file(
    volume_id: int,
    path: str,
):
    """Delete file in the volume.

    Args:
        volume_id(int): Volume ID.
        path(str): Path within the volume.

    Example:
        ```python
        vessl.delete_volume_file(
            volume_id=123456,
            path="model.pth",
        )
        ```
    """
    vessl_api.volume_file_delete_api(
        volume_id=volume_id,
        path=path,
    )


def upload_volume_file(volume_id: int, path: str) -> StorageFile:
    """Upload file in the volume.

    Args:
        volume_id(int): Volume ID.
        path(str): Local file path to upload

    Example:
        ```python
        vessl.upload_volume_file(
            volume_id=123456,
            path="model.pth",
        )
        ```
    """
    return vessl_api.volume_file_uploaded_api(volume_id=volume_id, path=path)


def copy_volume_file(
    source_volume_id: Optional[int],
    source_path: str,
    dest_volume_id: Optional[int],
    dest_path: str,
    quiet: bool = False,
) -> Optional[List[StorageFile]]:
    """Copy file either from local to remote, remote to local, or remote to
    remote.

    Args:
        source_volume_id(Optional[int]): Source volume file id. If not
            specified, source is assumed to be local.
        source_path(str): If source_volume_id is empty, local source path.
            Otherwise, remote source path.
        dest_volume_id(Optional[int]): Destination volume file id. If not
            specified, destination is assumed to be local.
        dest_path(str): If dest_volume_id is empty, local destination path.
            Otherwise, remote destination path.
        quiet(bool): True if the muted output, False otherwise. Defaults to
            False.

    Example:
        ```python
        vessl.copy_volume_file(
            source_volume_id=123456,
            source_path="model.pth",
            dest_volume_id=123457,
            dest_path="model.pth",
        )
        ```
    """
    if source_volume_id is None and dest_volume_id:
        return VolumeFileTransfer(dest_volume_id).upload(source_path, dest_path)

    if dest_volume_id is None and source_volume_id:
        VolumeFileTransfer(source_volume_id).download(source_path, dest_path)
        return

    if source_volume_id and dest_volume_id:
        return _copy_volume_file_remote_to_remote(
            source_volume_id,
            source_path,
            dest_volume_id,
            dest_path,
        )


def _copy_volume_file_local_to_remote(
    source_path: str, dest_volume_id: int, dest_path: str
) -> List[StorageFile]:
    """Copy local to remote

    Behavior works like linux cp command
    - `source_path` is file
      - `dest_path` is not a directory: copy as file with new name
      - `dest_path` is directory: copy file into directory with original name
    - `source_path` is directory
      - `dest_path` is file: error
      - `dest_path` does not exist: create `dest_path` and copy contents of `source_path`
      - `dest_path` exists: copy `source_path` as subdirectory of `dest_path`
    """

    output = "Successfully uploaded {} out of {} file(s)."
    source_path = source_path.rstrip("/")

    try:
        dest_file = read_volume_file(dest_volume_id, dest_path)
    except VesslApiException:
        dest_file = None

    if not os.path.isdir(source_path):
        if dest_file and (dest_file.is_dir or dest_path.endswith("/")):
            dest_path = os.path.join(dest_path, os.path.basename(source_path))

        uploaded_file = Uploader.upload(source_path, dest_volume_id, dest_path)

        print(output.format(1, 1))
        return [uploaded_file]

    if dest_file and not dest_file.is_dir:
        raise InvalidVolumeFileError(f"Destination path is not a directory: {dest_path}.")

    if dest_file and dest_file.is_dir:
        dest_path = os.path.join(dest_path, os.path.basename(source_path))

    paths = Uploader.get_paths_in_dir(source_path)
    uploaded_files = Uploader.bulk_upload(source_path, paths, dest_volume_id, dest_path)

    print(output.format(len(uploaded_files), len(paths)))
    return uploaded_files


def _copy_volume_file_remote_to_local(
    source_volume_id: int,
    source_path: str,
    dest_path: str,
    quiet: bool = False,
) -> None:
    """Copy remote to local

    Behavior works like linux cp command
    - `source_path` is file
      - `dest_path` is not a directory: copy as file with new name
      - `dest_path` is directory: copy file into directory with original name
    - `source_path` is directory
      - `dest_path` is file: error
      - `dest_path` does not exist: create `dest_path` and copy contents of `source_path` (like cp -r)
      - `dest_path` exists: copy `source_path` as subdirectory of `dest_path`
    """

    try:
        source_file = read_volume_file(source_volume_id, source_path)
    except VesslApiException:
        if os.path.isfile(dest_path):
            # Case where `source_path` is a directory and `dest_path` is an
            # existing filename
            raise InvalidVolumeFileError(f"Destination path is not a directory: {dest_path}.")

        files = list_volume_files(
            volume_id=source_volume_id,
            need_download_url=True,
            path=source_path,
            recursive=True,
        )

        Downloader.download(source_path, dest_path, *files, quiet=quiet)
        return

    file_name = os.path.basename(source_file.path)
    if os.path.isdir(dest_path):
        dest_path = os.path.join(dest_path, file_name)

        if os.path.isdir(dest_path):
            # Case where `source_path` is "a.txt", `dest_path` is "dir/", and
            # "dir/a.txt/" exists as a directory
            raise InvalidParamsError(
                f"Cannot overwrite directory {dest_path} with non-directory {file_name}."
            )

    Downloader.download(source_path, dest_path, source_file, quiet=quiet)


def _copy_volume_file_remote_to_remote(
    source_volume_id: int,
    source_path: str,
    dest_volume_id: int,
    dest_path: str,
) -> None:
    if source_volume_id != dest_volume_id:
        raise InvalidVolumeFileError("Files can only be copied within the same volume.")

    vessl_api.volume_file_copy_api(
        volume_id=source_volume_id,
        volume_file_copy_api_input=VolumeFileCopyAPIInput(
            dest_path=dest_path,
            source_dataset_version="latest",
            source_path=source_path,
        ),
    )

    print(f"Successfully copied.")


def _print_volume_mount_requests_tree(
    requests: OrmVolumeMountRequests,
):
    print("Mount Info:")
    for request in requests.requests:
        if request.source_type == SOURCE_TYPE_ARCHIVE_FILE:
            print(f"ㄴ {request.mount_path} -> (Uploaded)")
        if request.source_type == SOURCE_TYPE_CODE:
            print(
                f"ㄴ {request.mount_path} -> Code [{request.code.git_provider}@{request.code.git_owner}/{request.code.git_repo}@{request.code.git_ref}]"
            )
        if request.source_type == SOURCE_TYPE_DATASET:
            print(f"ㄴ {request.mount_path} -> Dataset [{request.dataset.dataset_name}]")
        if request.source_type == SOURCE_TYPE_DATASET_VERSION:
            print(
                f"ㄴ {request.mount_path} -> Dataset [{request.dataset_version.dataset_name}@{request.dataset_version.dataset_version_hash}]"
            )
        if request.source_type == SOURCE_TYPE_MODEL_VOLUME:
            print(
                f"ㄴ {request.mount_path} -> Model [{request.model_volume.model_repository_name}/{request.model_volume.model_number}{request.model_volume.sub_path}]"
            )
        if request.source_type == SOURCE_TYPE_OUTPUT:
            print(f"ㄴ {request.mount_path} -> (output)")
        if request.source_type == SOURCE_TYPE_OBJECT_STORAGE:
            print(f"ㄴ {request.mount_path} -> Object Storage({request.object_storage.mode}) [{request.object_storage.bucket_path}]")


def _configure_volume_mount_requests(
    dataset_mounts: Optional[List[str]],
    git_ref_mounts: Optional[List[str]],
    git_diff_mount: Optional[str],
    archive_file_mount: Optional[str],
    object_storage_mounts: Optional[List[str]],
    root_volume_size: Optional[str],
    working_dir: Optional[str],
    output_dir: Optional[str],
    model_mounts: Optional[List[str]] = None,
    local_files: Optional[List[str]] = None,
    upload_local_git_diff: bool = False,
    **kwargs,
) -> OrmVolumeMountRequests:
    requests = [
        _configure_volume_mount_request_empty_dir(),
        _configure_volume_mount_request_output(output_dir),
    ]

    if local_files:
        requests.extend(_configure_volume_mount_request_local_files(local_files, **kwargs))

    if not git_ref_mounts and git_diff_mount is None and archive_file_mount is None:
        # No information for code mount is given - user is creating new experiment from CLI.
        # Generate VolumeMountRequest from projectRepository and local working directory
        requests.extend(_configure_volume_mount_request_local_git(upload_local_git_diff, **kwargs))
    else:
        # Explicit information for code mount is given - user is reproducing experiment from CLI.
        if git_ref_mounts:
            requests.extend(
                _configure_volume_mount_request_codes(git_ref_mounts, git_diff_mount, **kwargs)
            )
        if archive_file_mount is not None:
            requests.append(
                _configure_volume_mount_request_archive_file(archive_file_mount, **kwargs)
            )

    if dataset_mounts is not None:
        requests.extend(_configure_volume_mount_request_datasets(dataset_mounts, **kwargs))

    if model_mounts is not None:
        requests += [
            _configure_volume_mount_request_model(model_mount, **kwargs)
            for model_mount in model_mounts
        ]

    if object_storage_mounts is not None:
        requests.extend(_configure_volume_mount_request_object_storages(object_storage_mounts))

    requests = OrmVolumeMountRequests(
        root_volume_size=root_volume_size,
        working_dir=working_dir,
        requests=requests,
    )
    _print_volume_mount_requests_tree(requests)
    return requests


def _configure_volume_mount_request_datasets(
    dataset_mounts: List[str], should_add_project_datasets=True, **kwargs
) -> List[OrmVolumeMountRequest]:
    from vessl.dataset import read_dataset, read_dataset_version

    dataset_mount_map: Dict[int, OrmVolumeMountRequest] = {}
    if should_add_project_datasets:
        project = _get_project(**kwargs)
        for project_dataset in project.project_datasets:
            mount_path = project_dataset.mount_path
            dataset = project_dataset.dataset

            dataset_mount_map[dataset.id] = OrmVolumeMountRequest(
                source_type=SOURCE_TYPE_DATASET,
                mount_path=mount_path,
                dataset=OrmVolumeMountRequestSourceDataset(
                    dataset_id=dataset.id,
                    dataset_name=dataset.name,
                ),
            )

    for dataset_mount in dataset_mounts:
        mount_path, dataset_path = dataset_mount.split(":", 1)
        mount_path = os.path.join(mount_path, "")  # Ensure path ends in /

        organization_name = _get_organization_name(**kwargs)
        dataset_name = dataset_path
        dataset_version_hash = DATASET_VERSION_HASH_LATEST

        if "@" in dataset_path:
            # Example: mnist@3d1e0f91c
            dataset_name, dataset_version_hash = dataset_path.rsplit("@", 1)

        if "/" in dataset_name:
            # Example: org1/mnist@3d1e0f91c
            organization_name, dataset_name = dataset_name.split("/", 1)

        dataset = read_dataset(dataset_name, organization_name=organization_name)

        if not dataset.is_version_enabled:
            dataset_mount_map[dataset.id] = OrmVolumeMountRequest(
                source_type=SOURCE_TYPE_DATASET,
                mount_path=mount_path,
                dataset=OrmVolumeMountRequestSourceDataset(
                    dataset_id=dataset.id,
                    dataset_name=dataset_name,
                ),
            )

        if dataset_version_hash != DATASET_VERSION_HASH_LATEST:
            dataset_version_hash = read_dataset_version(
                dataset.id, dataset_version_hash
            ).version_hash  # Get full version hash from truncated one given by input

            dataset_mount_map[dataset.id] = OrmVolumeMountRequest(
                source_type=SOURCE_TYPE_DATASET_VERSION,
                mount_path=mount_path,
                dataset_version=OrmVolumeMountRequestSourceDatasetVersion(
                    dataset_id=dataset.id,
                    dataset_name=dataset_name,
                    dataset_version_hash=dataset_version_hash,
                ),
            )

    return list(dataset_mount_map.values())


def _configure_volume_mount_request_model(model_mount: str, **kwargs) -> OrmVolumeMountRequest:
    from vessl.model import read_model

    mount_path, model_path = model_mount.split(":", 1)
    model_repository_name, model_number = model_path.split("/", 1)

    mount_path = os.path.join(mount_path, "")  # Ensure path ends in /

    model = read_model(repository_name=model_repository_name, model_number=model_number, **kwargs)

    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_MODEL_VOLUME,
        mount_path=mount_path,
        model_volume=OrmVolumeMountRequestSourceModelVolume(
            sub_path="/",
            model_repository_name=model.model_repository.name,
            model_number=model.number,
        ),
    )


def _configure_volume_mount_request_local_files(
    local_files: List[str], **kwargs
) -> List[OrmVolumeMountRequest]:
    project = _get_project(**kwargs)

    volume_mount_requests: List[OrmVolumeMountRequest] = []
    for local_file in local_files:
        if ":" in local_file:
            local_path, remote_path = local_file.split(":")
            if not remote_path.startswith("/"):
                remote_path = f"{EXPERIMENT_WORKING_DIR}{remote_path}"
        else:
            local_path = local_file
            remote_path = f"{EXPERIMENT_WORKING_DIR}local"

        cwd = os.path.abspath(local_path)
        gzip_file_full_path = os.path.abspath(
            os.path.join(TEMP_DIR, f"{project.name}_{random_string()}.tar.gz")
        )

        print(f"Uploading... [{cwd} -> {remote_path}]")
        Tar.gzip(gzip_file_full_path, cwd)

        uploaded = Uploader.upload(
            local_path=gzip_file_full_path,
            volume_id=project.volume_id,
            remote_path=os.path.basename(gzip_file_full_path),
        )
        print(f"Upload completed successfully.")
        os.remove(gzip_file_full_path)

        volume_mount_requests.append(
            _configure_volume_mount_request_archive_file(
                archive_file_mount=f"{remote_path}:{uploaded.path}"
            )
        )

    return volume_mount_requests


def _configure_volume_mount_request_local_git(
    upload_local_git_diff, **kwargs
) -> List[OrmVolumeMountRequest]:
    project = _get_project(**kwargs)

    local_git_owner, local_git_repo = None, None
    try:
        local_git_owner, local_git_repo = get_git_repo()
    except GitError:
        pass

    volume_mount_requests: List[OrmVolumeMountRequest] = []
    used_mount_paths = set()
    for repo in project.project_repositories:
        mount_path = f"{EXPERIMENT_WORKING_DIR}{repo.git_repo}"
        if mount_path in used_mount_paths:
            mount_path = f"{EXPERIMENT_WORKING_DIR}{repo.git_repo}-{repo.git_owner}"

        git_mount_info = f"github/{repo.git_owner}/{repo.git_repo}/latest"
        git_diff_path = None
        if (
            upload_local_git_diff
            and repo.git_owner == local_git_owner
            and repo.git_repo == local_git_repo
        ):
            # Current working directory is one of the project repository.
            # Mount VolumeSourceCode with potential uncommitted changes
            git_diff_path = get_git_diff_path(project)
            git_mount_info = git_mount_info.replace("latest", get_git_ref())
        volume_mount_requests.append(
            _configure_volume_mount_request_code(
                git_ref_mount=f"{mount_path}:{git_mount_info}",
                git_diff_mount=None if git_diff_path is None else f"{mount_path}:{git_diff_path}",
            )
        )

    return volume_mount_requests


def _configure_volume_mount_request_codes(
    git_ref_mounts: List[str], git_diff_mount: str, **kwargs
) -> List[OrmVolumeMountRequest]:
    return [
        _configure_volume_mount_request_code(git_ref_mount, git_diff_mount, **kwargs)
        for git_ref_mount in git_ref_mounts
    ]


def _configure_volume_mount_request_code(
    git_ref_mount: str, git_diff_mount: str, **kwargs
) -> OrmVolumeMountRequest:
    # Generate OrmVolumeMountRequestSourceCode from git_ref_mount
    # ref_mount_info = <provider>/<owner>/<repo>/<commit>
    ref_mount_path, ref_mount_info = git_ref_mount.split(":", 1)
    code_info = ref_mount_info.split("/", 3)

    vmr_source_code = OrmVolumeMountRequestSourceCode(
        git_provider=code_info[0],
        git_owner=code_info[1],
        git_repo=code_info[2],
        git_ref=code_info[3],
    )

    # Add optional git_diff parameter if git_diff_mount has same mount path
    if git_diff_mount is not None:
        diff_mount_path, diff_mount_file = git_diff_mount.split(":", 1)
        if ref_mount_path == diff_mount_path:
            vmr_source_code.git_diff_file = diff_mount_file

    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_CODE,
        mount_path=ref_mount_path,
        code=vmr_source_code,
    )


def _configure_volume_mount_request_archive_file(
    archive_file_mount: str,
) -> OrmVolumeMountRequest:
    mount_path, archive_file_path = archive_file_mount.split(":", 1)

    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_ARCHIVE_FILE,
        mount_path=mount_path,
        archive_file=OrmVolumeMountRequestSourceArchiveFile(
            archive_file=archive_file_path,
        ),
    )

def _configure_volume_mount_request_object_storage(
    object_storage_mount: str,
) -> OrmVolumeMountRequest:
    mode, path = object_storage_mount.split("@", 1)
    mount_path, bucket_path = path.split(":", 1)

    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_OBJECT_STORAGE,
        mount_path=mount_path,
        object_storage=OrmVolumeMountRequestSourceObjectStorage(
            mode=mode,
            bucket_path=bucket_path,
        ),
    )

def _configure_volume_mount_request_object_storages(
    object_storage_mounts: List[str], **kwargs
) -> List[OrmVolumeMountRequest]:
    return [
        _configure_volume_mount_request_object_storage(object_storage_mount, **kwargs)
        for object_storage_mount in object_storage_mounts
    ]

def _configure_volume_mount_request_empty_dir() -> OrmVolumeMountRequest:
    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_EMPTY_DIR,
        mount_path=MOUNT_PATH_EMPTY_DIR,
    )


def _configure_volume_mount_request_output(output_dir: str) -> OrmVolumeMountRequest:
    return OrmVolumeMountRequest(
        source_type=SOURCE_TYPE_OUTPUT,
        mount_path=output_dir,
    )



import os.path
import subprocess
from typing import List


class Tar:
    @staticmethod
    def gzip(output_file_name: str, path_to_zip: str, exclude_paths: List = ()) -> int:
        """
        exclude_paths: relative path from path_to_zip
        """
        commands = ["tar", "-C", path_to_zip]
        for exclude_path in exclude_paths:
            commands.extend(["--exclude", exclude_path])
        commands.extend(["-zcf", output_file_name, "."])

        subprocess.run(commands).check_returncode()
        return os.path.getsize(output_file_name)

    @staticmethod
    def extract(tar_file_name: str, path_to_extract: str):
        subprocess.run(["mkdir", "-p", path_to_extract]).check_returncode()
        subprocess.run(["tar", "xf", tar_file_name, "-C", path_to_extract]).check_returncode()
